from urllib.parse import urlencode


class QueryStringTransformer:
    """
    A utility class to manipulate query strings.

    Args:
        args: An object representing the query parameters, typically an
              ImmutableMultiDict (Django) or QueryDict (Flask) which can be
              accessed with request.GET (Django) or request.args (Flask).
    """

    def __init__(self, args):
        try:
            args_lists = args.lists()
        except AttributeError:
            raise TypeError(
                "args must be a ImmutableMultiDict (Django) or QueryDict (Flask) object"
            )
        self.args = list(args_lists)

    def parameter_exists(self, parameter) -> bool:
        """
        Check if a parameter exists in the query parameters.
        """

        return parameter in dict(self.args)

    def parameter_values(self, parameter) -> list:
        """
        Get the values associated with a parameter in the query parameters.
        Raises an AttributeError if the parameter does not exist.
        """

        for key, values in self.args:
            if key == parameter:
                return values
        raise AttributeError(f"Parameter '{parameter}' does not exist")

    def add_parameter(self, parameter, values=None) -> "QueryStringTransformer":
        """
        Add a new parameter to the query parameters.
        Raises a ValueError if the parameter already exists.
        """

        for key, vals in self.args:
            if key == parameter:
                raise ValueError(f"Parameter '{parameter}' already exists")
        if type(values) is not list:
            values = [str(values)] if values is not None else []
        else:
            values = [str(v) for v in values]
        self.args.append((parameter, values))
        return self

    def update_parameter(self, parameter, values=None) -> "QueryStringTransformer":
        """
        Update an existing parameter in the query parameters.
        If the parameter does not exist, it will be added.
        """

        try:
            self.remove_parameter(parameter)
        except AttributeError:
            pass
        self.add_parameter(parameter, values)
        return self

    def remove_parameter(self, parameter) -> "QueryStringTransformer":
        """
        Remove a parameter from the query parameters.
        Raises an AttributeError if the parameter does not exist.
        """

        for key, vals in self.args:
            if key == parameter:
                self.args.remove((key, vals))
                return self
        raise AttributeError(f"Parameter '{parameter}' does not exist")

    def is_value_in_parameter(self, parameter, value) -> bool:
        """
        Check if a specific value exists within a parameter's values.
        Raises an AttributeError if the parameter does not exist.
        """

        for key, values in self.args:
            if key == parameter:
                return value in values
        raise AttributeError(f"Parameter '{parameter}' does not exist")

    def add_parameter_value(self, parameter, value) -> "QueryStringTransformer":
        """
        Add a specific value to a parameter's values.
        Raises an AttributeError if the parameter does not exist.
        """

        for key, values in self.args:
            if key == parameter:
                if value not in values:
                    values.append(value)
                return self
        raise AttributeError(f"Parameter '{parameter}' does not exist")

    def toggle_parameter_value(self, parameter, value) -> "QueryStringTransformer":
        """
        Toggle a value within a parameter's values.
        If the value exists, it will be removed; if it does not exist, it will be added.
        Raises an AttributeError if the parameter does not exist.
        """

        for key, values in self.args:
            if key == parameter:
                if value in values:
                    self.remove_parameter_value(parameter, value)
                else:
                    self.add_parameter_value(parameter, value)
                return self
        raise AttributeError(f"Parameter '{parameter}' does not exist")

    def remove_parameter_value(self, parameter, value) -> "QueryStringTransformer":
        """
        Remove a specific value from a parameter's values.
        Raises an AttributeError if the parameter does not exist.
        """

        for key, values in self.args:
            if key == parameter:
                if value in values:
                    values.remove(value)
                return self
        raise AttributeError(f"Parameter '{parameter}' does not exist")

    def get_query_string(self) -> str:
        """
        Get the full query string.
        """

        return f"?{urlencode(self.args, doseq=True)}"



import unittest

from tna_utilities import strtobool


class TestStrToBool(unittest.TestCase):
    def test_happy_truthy(self):
        self.assertEqual(strtobool("y"), True)
        self.assertEqual(strtobool("Y"), True)
        self.assertEqual(strtobool("t"), True)
        self.assertEqual(strtobool("true"), True)
        self.assertEqual(strtobool("on"), True)
        self.assertEqual(strtobool("1"), True)

    def test_happy_falsy(self):
        self.assertEqual(strtobool("n"), False)
        self.assertEqual(strtobool("N"), False)
        self.assertEqual(strtobool("no"), False)
        self.assertEqual(strtobool("f"), False)
        self.assertEqual(strtobool("false"), False)
        self.assertEqual(strtobool("off"), False)
        self.assertEqual(strtobool("0"), False)

    def test_unhappy_invalid(self):
        with self.assertRaises(ValueError):
            strtobool("")
        with self.assertRaises(ValueError):
            strtobool("maybe")
        with self.assertRaises(TypeError):
            strtobool(True)
        with self.assertRaises(TypeError):
            strtobool(False)
        with self.assertRaises(TypeError):
            strtobool(42)
        with self.assertRaises(TypeError):
            strtobool(None)
        with self.assertRaises(TypeError):
            strtobool({})
        with self.assertRaises(TypeError):
            strtobool(())



from typing import Optional, Union


def currency(value: Union[float, str, int], simplify: bool = True) -> str:
    """
    Formats a number as a currency without the currency symbol.

    If simplify is True, removes unnecessary decimal places for whole numbers.
    """
    if not value:
        return "0" if simplify else "0.00"

    float_number = float(value)
    int_number = int(float_number)

    if simplify and int_number == float_number:
        return str("{:,}".format(int_number))

    return str("{:,.2f}".format(float_number))


def pretty_price(
    value: Union[float, str, int],
    simplify: bool = True,
    currency_symbol: str = "£",
) -> str:
    """
    Formats a number as a price.

    If the value is 0, returns "Free".
    Otherwise, returns the currency symbol followed by the formatted currency.
    """
    if value == 0 or value == "0" or round(float(value) * 100) == 0:
        return "Free"

    return f"{currency_symbol}{currency(value, simplify)}"


def pretty_price_range(
    value_from: Optional[Union[float, str, int]] = None,
    value_to: Optional[Union[float, str, int]] = None,
    simplify: bool = True,
    currency_symbol: str = "£",
) -> str:
    """
    Formats a price range.

    If both value_from and value_to are 0 or None, returns "Free".
    If value_from equals value_to, returns the pretty price of that value.
    If value_from is 0 or None, returns "Free to {value_to}".
    If value_to is 0 or None, returns "From {value_from}".
    Otherwise, returns "{min_price} to {max_price}".
    """

    if value_from is not None and (
        not isinstance(value_from, (str, float, int)) or type(value_from) is bool
    ):
        raise TypeError("value_from must be a string, float, int, or None")

    if value_to is not None and (
        not isinstance(value_to, (str, float, int)) or type(value_to) is bool
    ):
        raise TypeError("value_to must be a string, float, int, or None")

    value_from = float(value_from) if value_from is not None else 0
    value_to = float(value_to) if value_to is not None else 0

    if value_from == 0 and value_to == 0:
        return "Free"

    if value_from == value_to:
        return pretty_price(value_from, simplify, currency_symbol)

    if value_from == 0:
        return f"Free to {pretty_price(value_to, simplify, currency_symbol)}"

    if value_to == 0:
        return f"From {pretty_price(value_from, simplify, currency_symbol)}"

    (min_price, max_price) = sorted([float(value_from), float(value_to)])
    return f"{pretty_price(min_price, simplify, currency_symbol)} to {pretty_price(max_price, simplify, currency_symbol)}"



import datetime
import math
from typing import Optional, Union

"""
See https://design-system.nationalarchives.gov.uk/content/dates-and-times/
for more details on how dates and times should be presented within services
from The National Archives.
"""


def get_date_from_string(date_string: str) -> datetime.datetime:  # noqa: C901
    """
    Parses a date string into a datetime object.
    """

    if not date_string:
        raise ValueError("Empty string cannot be parsed as date")

    s = date_string.replace("Z", "+00:00")

    try:
        return datetime.datetime.fromisoformat(s)
    except ValueError:
        pass

    try:
        return datetime.datetime.strptime(s, "%Y-%m-%dT%H:%M:%S.%fZ")
    except ValueError:
        pass

    try:
        return datetime.datetime.strptime(s, "%Y-%m-%dT%H:%M:%SZ")
    except ValueError:
        pass

    try:
        return datetime.datetime.strptime(s, "%Y-%m-%dT%H:%M:%S%z")
    except ValueError:
        pass

    try:
        return datetime.datetime.strptime(s, "%Y-%m-%d")
    except ValueError:
        pass

    try:
        return datetime.datetime.strptime(s, "%Y-%m")
    except ValueError:
        pass

    try:
        return datetime.datetime.strptime(s, "%Y")
    except ValueError:
        pass

    raise ValueError(f"Unable to parse date from string: {date_string}")


def pretty_date(
    date: Union[str, datetime.date, datetime.datetime],
    show_day: bool = False,
) -> str:
    """
    Formats a date into the format used by The National Archives.

    date: A date or datetime object or the date string in either YYYY-MM-DD, YYYY-MM, or YYYY format.

    If show_day is True, includes the day of the week.
    """

    if not date:
        raise ValueError("No date provided")

    if isinstance(date, datetime.date):
        if show_day:
            return date.strftime("%A %-d %B %Y")
        return date.strftime("%-d %B %Y")

    try:
        date = datetime.datetime.strptime(date, "%Y-%m-%d")
        if show_day:
            return date.strftime("%A %-d %B %Y")
        return date.strftime("%-d %B %Y")
    except ValueError:
        pass

    try:
        date = datetime.datetime.strptime(date, "%Y-%m")
        return date.strftime("%B %Y")
    except ValueError:
        pass

    try:
        date = datetime.datetime.strptime(date, "%Y")
        return date.strftime("%Y")
    except ValueError:
        pass

    date = get_date_from_string(date)
    if show_day:
        return date.strftime("%A %-d %B %Y")
    return date.strftime("%-d %B %Y")


def pretty_datetime(
    date: Union[str, datetime.datetime],
    show_day: bool = False,
) -> str:
    """
    Formats a date and time into the format used by The National Archives.
    """

    if not date:
        raise ValueError("No date provided")

    if isinstance(date, datetime.datetime):
        if show_day:
            return date.strftime("%A %-d %B %Y, %H:%M")
        return date.strftime("%-d %B %Y, %H:%M")

    if isinstance(date, datetime.date):
        raise TypeError("Date object provided, datetime object expected")

    date = get_date_from_string(date)
    if show_day:
        return date.strftime("%A %-d %B %Y, %H:%M")
    return date.strftime("%-d %B %Y, %H:%M")


def pretty_date_range(  # noqa: C901
    date_from: Optional[Union[str, datetime.datetime]],
    date_to: Optional[Union[str, datetime.datetime]],
    omit_days: bool = False,
    lowercase_first: bool = False,
) -> str:
    """
    Formats a date range into the format used by The National Archives.
    """

    if date_from and isinstance(date_from, (datetime.date, datetime.datetime)):
        date_from = date_from
    else:
        try:
            date_from = get_date_from_string(date_from)
        except ValueError:
            pass

    if date_to and isinstance(date_to, (datetime.date, datetime.datetime)):
        date_to = date_to
    else:
        try:
            date_to = get_date_from_string(date_to)
        except ValueError:
            pass

    if not date_from and not date_to:
        raise ValueError("No dates provided")

    if date_from and date_to:
        if date_from > date_to:
            raise ValueError("From date is after to date")

        date_to_string = date_to.strftime("%B %Y" if omit_days else ("%-d %B %Y"))
        if (
            date_from.day == 1
            and date_from.month == 1
            and date_to.day == 31
            and date_to.month == 12
        ):
            if date_from.year == date_to.year:
                return str(date_from.year)

            return f"{date_from.year} to {date_to.year}"

        if date_from.year == date_to.year:
            if date_from.month == date_to.month:
                if date_from.day == date_to.day:
                    return date_from.strftime("%B %Y" if omit_days else "%-d %B %Y")

                if omit_days:
                    return date_to_string
                return f"{date_from.strftime('%-d')} to {date_to_string}"

            return f"{date_from.strftime('%B' if omit_days else '%-d %B')} to {date_to_string}"

        return f"{date_from.strftime('%B %Y' if omit_days else '%-d %B %Y')} to {date_to_string}"

    if date_from:
        start = "from" if lowercase_first else "From"
        return f"{start} {date_from.strftime('%B %Y' if omit_days else '%-d %B %Y')}"

    if date_to:
        start = "now to" if lowercase_first else "Now to"
        return f"{start} {date_to.strftime('%B %Y' if omit_days else '%-d %B %Y')}"


def pretty_datetime_range(  # noqa: C901
    date_from: Optional[Union[str, datetime.date, datetime.datetime]],
    date_to: Optional[Union[str, datetime.date, datetime.datetime]],
    lowercase_first: bool = False,
    hide_date_if_single_day: bool = False,
) -> str:
    """
    Formats a date/time range into the format used by The National Archives.
    """

    if date_from and isinstance(date_from, datetime.datetime):
        date_from = date_from
    elif date_from and isinstance(date_from, datetime.date):
        raise TypeError("From date object provided, datetime object expected")
    else:
        try:
            date_from = get_date_from_string(date_from)
        except ValueError:
            pass

    if date_to and isinstance(date_to, datetime.datetime):
        date_to = date_to
    elif date_to and isinstance(date_to, datetime.date):
        raise TypeError("To date object provided, datetime object expected")
    else:
        try:
            date_to = get_date_from_string(date_to)
        except ValueError:
            pass

    if not date_from and not date_to:
        raise ValueError("No dates provided")

    if date_from and date_to:
        if date_from > date_to:
            raise ValueError("From date is after to date")

        if (
            date_from.year == date_to.year
            and date_from.month == date_to.month
            and date_from.day == date_to.day
        ):
            if date_from.hour != date_to.hour or date_from.minute != date_to.minute:
                if hide_date_if_single_day:
                    return (
                        f"{date_from.strftime('%H:%M')} to {date_to.strftime('%H:%M')}"
                    )

                return f"{date_from.strftime('%-d %B %Y, %H:%M')} to {date_to.strftime('%H:%M')}"

            if hide_date_if_single_day:
                return f"{date_from.strftime('%H:%M')}"

            return f"{date_from.strftime('%-d %B %Y, %H:%M')}"

        return f"{date_from.strftime('%-d %B %Y, %H:%M')} to {date_to.strftime('%-d %B %Y, %H:%M')}"

    if date_from:
        start = "from" if lowercase_first else "From"
        return f"{start} {date_from.strftime('%-d %B %Y, %H:%M')}"

    if date_to:
        start = "now to" if lowercase_first else "Now to"
        return f"{start} {date_to.strftime('%-d %B %Y, %H:%M')}"


def is_today_or_future(date: Union[datetime.date, datetime.datetime]) -> bool:
    """
    Determines if the given date string represents today or a future date.
    """

    if not date:
        raise ValueError("No date provided")

    if isinstance(date, datetime.datetime):
        date = date.date()

    today = datetime.datetime.now().date()
    return today <= date


def is_today_in_date_range(
    date_from: Union[datetime.date, datetime.datetime],
    date_to: Union[datetime.date, datetime.datetime],
) -> bool:
    """
    Determines if today's date falls within the given date range.
    """

    if not date_from or not date_to:
        raise ValueError("Both from and to dates must be provided")

    if isinstance(date_from, datetime.datetime):
        date_from = date_from.date()

    if isinstance(date_to, datetime.datetime):
        date_to = date_to.date()

    today = datetime.datetime.now().date()
    return date_from <= today <= date_to


def group_by_year_and_month(
    items: list[dict], date_key: str, reverse: bool = False
) -> dict:  # noqa: C901
    """
    Groups a list of items by year and month based on a date key in each item.
    """

    grouped = []

    for item in items.get("items", []):
        if request_date := item.get(date_key):
            if isinstance(request_date, (datetime.datetime, datetime.date)):
                request_datetime = request_date
            else:
                try:
                    request_datetime = get_date_from_string(request_date)
                except ValueError:
                    continue

            if request_datetime:
                month = request_datetime.strftime("%B")
                month_index = int(request_datetime.strftime("%-m"))
                year = request_datetime.strftime("%Y")
                year_index = int(year)
                existing_year_index = next(
                    (i for i, d in enumerate(grouped) if d["index"] == year_index), None
                )
                if existing_year_index is None:
                    grouped.append(
                        {
                            "heading": year,
                            "index": year_index,
                            "items": [
                                {
                                    "heading": month,
                                    "index": month_index,
                                    "items": [item],
                                }
                            ],
                        }
                    )
                else:
                    existing_month_index = next(
                        (
                            i
                            for i, m in enumerate(grouped[existing_year_index]["items"])
                            if m["index"] == month_index
                        ),
                        None,
                    )
                    if existing_month_index is None:
                        grouped[existing_year_index]["items"].append(
                            {
                                "heading": month,
                                "index": month_index,
                                "items": [item],
                            }
                        )
                    else:
                        grouped[existing_year_index]["items"][existing_month_index][
                            "items"
                        ].append(item)

    for year_group in grouped:
        year_group["items"].sort(key=lambda x: x["index"], reverse=reverse)
        for month_group in year_group["items"]:
            month_group["items"].sort(
                key=lambda x: (
                    x.get(date_key)
                    if isinstance(x.get(date_key), (datetime.date, datetime.datetime))
                    else get_date_from_string(x.get(date_key))
                ),
                reverse=reverse,
            )
    grouped.sort(key=lambda x: x["index"], reverse=reverse)

    return grouped


def seconds_to_iso_8601_duration(total_seconds: int) -> str:
    """
    Converts a total number of seconds into an ISO 8601 duration string.
    """

    if not total_seconds:
        return "PT0S"

    if total_seconds < 0:
        raise ValueError("Total seconds cannot be negative")

    hours = math.floor(total_seconds / 3600)
    minutes = math.floor((total_seconds - (hours * 3600)) / 60)
    seconds = total_seconds - (hours * 3600) - (minutes * 60)
    if hours:
        return f"PT{hours}H{minutes}M{seconds}S"

    if minutes:
        return f"PT{minutes}M{seconds}S"

    return f"PT{seconds}S"


def seconds_to_duration(total_seconds: int, simplify: bool = False) -> str:
    """
    Converts a total number of seconds into a human-readable duration string.
    """

    if not total_seconds:
        return "00s" if simplify else "00h 00m 00s"

    if total_seconds < 0:
        raise ValueError("Total seconds cannot be negative")

    hours = math.floor(total_seconds / 3600)
    minutes = math.floor((total_seconds - (hours * 3600)) / 60)
    seconds = total_seconds - (hours * 3600) - (minutes * 60)

    hours_string = f"{str(hours).rjust(2, '0')}h"
    minutes_string = f"{str(minutes).rjust(2, '0')}m"
    seconds_string = f"{str(seconds).rjust(2, '0')}s"

    return_values = []
    if simplify:
        if hours:
            return_values.append(hours_string)
            return_values.append(minutes_string)
        elif minutes:
            return_values.append(minutes_string)
    else:
        return_values.append(hours_string)
        return_values.append(minutes_string)
    return_values.append(seconds_string)

    return " ".join(return_values)


def rfc_822_date_format(date: Union[datetime.date, datetime.datetime]) -> str:
    """
    Formats a date into RFC 822 format.
    """

    if not date:
        raise ValueError("No date provided")

    return date.strftime("%a, %-d %b %Y %H:%M:%S GMT")



import unittest

from tna_utilities.url import QueryStringTransformer


class TestQueryStringObject:
    def lists(self):
        return iter(
            [
                ("a", ["1"]),
                ("b", ["2", "3"]),
            ]
        )


class TestQuery(unittest.TestCase):
    def test_init(self):
        test_query = TestQueryStringObject()
        manipulator = QueryStringTransformer(test_query)
        self.assertEqual(manipulator.get_query_string(), "?a=1&b=2&b=3")

    def test_parameter_values(self):
        test_query = TestQueryStringObject()
        manipulator = QueryStringTransformer(test_query)
        self.assertEqual(manipulator.parameter_values("a"), ["1"])
        self.assertEqual(manipulator.parameter_values("b"), ["2", "3"])
        with self.assertRaises(AttributeError):
            manipulator.parameter_values("c")

    def test_add_parameter(self):
        test_query = TestQueryStringObject()
        manipulator = QueryStringTransformer(test_query)

        self.assertEqual(manipulator.add_parameter("c", []), manipulator)
        self.assertTrue(manipulator.parameter_exists("c"))
        self.assertEqual(manipulator.parameter_values("c"), [])

        self.assertEqual(manipulator.add_parameter("d", None), manipulator)
        self.assertTrue(manipulator.parameter_exists("d"))
        self.assertEqual(manipulator.parameter_values("d"), [])

        self.assertEqual(manipulator.add_parameter("e", ""), manipulator)
        self.assertTrue(manipulator.parameter_exists("e"))
        self.assertEqual(manipulator.parameter_values("e"), [""])

        self.assertEqual(manipulator.add_parameter("f", "4"), manipulator)
        self.assertTrue(manipulator.parameter_exists("f"))
        self.assertEqual(manipulator.parameter_values("f"), ["4"])

        self.assertEqual(manipulator.add_parameter("g", ["5", "6"]), manipulator)
        self.assertTrue(manipulator.parameter_exists("g"))
        self.assertEqual(manipulator.parameter_values("g"), ["5", "6"])

        self.assertEqual(manipulator.add_parameter("h", [False]), manipulator)
        self.assertTrue(manipulator.parameter_exists("h"))
        self.assertEqual(manipulator.parameter_values("h"), ["False"])

        self.assertEqual(
            manipulator.get_query_string(), "?a=1&b=2&b=3&e=&f=4&g=5&g=6&h=False"
        )

    def test_update_parameter(self):
        test_query = TestQueryStringObject()
        manipulator = QueryStringTransformer(test_query)
        self.assertEqual(manipulator.update_parameter("a", "10"), manipulator)
        self.assertEqual(manipulator.parameter_values("a"), ["10"])
        self.assertEqual(manipulator.update_parameter("b", ["20", "30"]), manipulator)
        self.assertEqual(manipulator.parameter_values("b"), ["20", "30"])
        self.assertEqual(manipulator.update_parameter("c", ["40"]), manipulator)
        self.assertEqual(manipulator.parameter_values("c"), ["40"])
        self.assertEqual(manipulator.get_query_string(), "?a=10&b=20&b=30&c=40")

    def test_remove_parameter(self):
        test_query = TestQueryStringObject()
        manipulator = QueryStringTransformer(test_query)
        self.assertEqual(manipulator.remove_parameter("a"), manipulator)
        self.assertFalse(manipulator.parameter_exists("a"))
        self.assertEqual(manipulator.remove_parameter("b"), manipulator)
        self.assertFalse(manipulator.parameter_exists("b"))
        with self.assertRaises(AttributeError):
            self.assertEqual(manipulator.remove_parameter("c"), manipulator)
        self.assertEqual(manipulator.get_query_string(), "?")

    def test_is_value_in_parameter(self):
        test_query = TestQueryStringObject()
        manipulator = QueryStringTransformer(test_query)
        self.assertTrue(manipulator.is_value_in_parameter("a", "1"))
        self.assertTrue(manipulator.is_value_in_parameter("b", "2"))
        self.assertTrue(manipulator.is_value_in_parameter("b", "3"))
        self.assertFalse(manipulator.is_value_in_parameter("b", "4"))
        with self.assertRaises(AttributeError):
            self.assertFalse(manipulator.is_value_in_parameter("c", "5"))

    def test_toggle_parameter_value(self):
        test_query = TestQueryStringObject()
        manipulator = QueryStringTransformer(test_query)
        self.assertEqual(manipulator.toggle_parameter_value("a", "1"), manipulator)
        self.assertFalse(manipulator.is_value_in_parameter("a", "1"))
        self.assertEqual(manipulator.toggle_parameter_value("a", "10"), manipulator)
        self.assertTrue(manipulator.is_value_in_parameter("a", "10"))
        self.assertEqual(manipulator.toggle_parameter_value("b", "2"), manipulator)
        self.assertFalse(manipulator.is_value_in_parameter("b", "2"))
        self.assertEqual(manipulator.get_query_string(), "?a=10&b=3")
        self.assertEqual(manipulator.toggle_parameter_value("a", "1"), manipulator)
        self.assertTrue(manipulator.is_value_in_parameter("a", "1"))
        self.assertEqual(manipulator.get_query_string(), "?a=10&a=1&b=3")

    def test_add_remove_parameter_value(self):
        test_query = TestQueryStringObject()
        manipulator = QueryStringTransformer(test_query)
        self.assertEqual(manipulator.add_parameter_value("a", "10"), manipulator)
        self.assertTrue(manipulator.is_value_in_parameter("a", "10"))
        self.assertEqual(manipulator.parameter_values("a"), ["1", "10"])

    def test_remove_parameter_value(self):
        test_query = TestQueryStringObject()
        manipulator = QueryStringTransformer(test_query)
        self.assertEqual(manipulator.remove_parameter_value("b", "2"), manipulator)
        self.assertFalse(manipulator.is_value_in_parameter("b", "2"))
        self.assertEqual(manipulator.parameter_values("b"), ["3"])



import unittest

from tna_utilities.string import slugify, unslugify


class TestSlugify(unittest.TestCase):
    def test_happy(self):
        self.assertEqual(slugify(""), "")
        self.assertEqual(slugify("test"), "test")
        self.assertEqual(slugify("  test TEST"), "test-test")
        self.assertEqual(slugify("test 12 3 -4 "), "test-12-3-4")
        self.assertEqual(slugify("test---test"), "test-test")
        self.assertEqual(slugify("test---"), "test")
        self.assertEqual(slugify("test---$"), "test")
        self.assertEqual(slugify("test---$---"), "test")

    def test_happy_unicode(self):
        self.assertEqual(slugify("tést"), "test")
        self.assertEqual(slugify("téßt"), "tet")
        self.assertEqual(slugify("tést", allow_unicode=True), "tést")
        self.assertEqual(slugify("téßt", allow_unicode=True), "téßt")

    def test_unhappy_none(self):
        with self.assertRaises(TypeError):
            slugify(None)

    def test_unhappy_boolean(self):
        with self.assertRaises(TypeError):
            slugify(False)

    def test_unhappy_list(self):
        with self.assertRaises(TypeError):
            slugify([])

    def test_unhappy_tuple(self):
        with self.assertRaises(TypeError):
            slugify(("a", "b"))

    def test_unhappy_int(self):
        with self.assertRaises(TypeError):
            slugify(123)

    def test_unhappy_dict(self):
        with self.assertRaises(TypeError):
            slugify({})


class TestUnslugify(unittest.TestCase):
    def test_happy(self):
        self.assertEqual(unslugify("test-test"), "Test test")
        self.assertEqual(unslugify("test-test", capitalize_first=False), "test test")
        self.assertEqual(unslugify("test-123"), "Test 123")
        self.assertEqual(unslugify("test-1-2-3"), "Test 1 2 3")

    def test_unhappy_none(self):
        with self.assertRaises(TypeError):
            unslugify(None)

    def test_unhappy_boolean(self):
        with self.assertRaises(TypeError):
            unslugify(False)

    def test_unhappy_list(self):
        with self.assertRaises(TypeError):
            unslugify([])

    def test_unhappy_tuple(self):
        with self.assertRaises(TypeError):
            unslugify(("a", "b"))

    def test_unhappy_int(self):
        with self.assertRaises(TypeError):
            unslugify(123)

    def test_unhappy_dict(self):
        with self.assertRaises(TypeError):
            unslugify({})



import unittest

from tna_utilities.currency import currency, pretty_price, pretty_price_range


class TestCurrency(unittest.TestCase):
    def test_happy_number(self):
        self.assertEqual(currency(0), "0")
        self.assertEqual(currency(5), "5")
        self.assertEqual(currency(5.0), "5")
        self.assertEqual(currency(5.00), "5")
        self.assertEqual(currency(5.1), "5.10")
        self.assertEqual(currency(5.01), "5.01")
        self.assertEqual(currency(5.001), "5.00")
        self.assertEqual(currency(5.005), "5.00")
        self.assertEqual(currency(5.006), "5.01")

    def test_happy_string(self):
        self.assertEqual(currency("0"), "0")
        self.assertEqual(currency("5"), "5")
        self.assertEqual(currency("5.0"), "5")
        self.assertEqual(currency("5.00"), "5")
        self.assertEqual(currency("5.1"), "5.10")
        self.assertEqual(currency("5.01"), "5.01")
        self.assertEqual(currency("5.001"), "5.00")
        self.assertEqual(currency("5.005"), "5.00")
        self.assertEqual(currency("5.006"), "5.01")

    def test_happy_not_simplified(self):
        self.assertEqual(currency("0", False), "0.00")
        self.assertEqual(currency("5", False), "5.00")
        self.assertEqual(currency("5.0", False), "5.00")
        self.assertEqual(currency("5.00", False), "5.00")


class TestPrettyPrice(unittest.TestCase):
    def test_happy_number(self):
        self.assertEqual(pretty_price(0), "Free")
        self.assertEqual(pretty_price(0.1), "£0.10")
        self.assertEqual(pretty_price(0.101), "£0.10")
        self.assertEqual(pretty_price(0.001), "Free")
        self.assertEqual(pretty_price(0.009), "£0.01")
        self.assertEqual(pretty_price(1), "£1")
        self.assertEqual(pretty_price(1.1), "£1.10")
        self.assertEqual(pretty_price(1.11), "£1.11")
        self.assertEqual(pretty_price(1.111), "£1.11")
        self.assertEqual(pretty_price(123456789), "£123,456,789")
        self.assertEqual(pretty_price(123456789.01), "£123,456,789.01")

    def test_happy_string(self):
        self.assertEqual(pretty_price("0"), "Free")
        self.assertEqual(pretty_price("0.1"), "£0.10")
        self.assertEqual(pretty_price("0.10"), "£0.10")
        self.assertEqual(pretty_price("0.101"), "£0.10")
        self.assertEqual(pretty_price("0.001"), "Free")
        self.assertEqual(pretty_price("0.009"), "£0.01")
        self.assertEqual(pretty_price("1"), "£1")
        self.assertEqual(pretty_price("01"), "£1")
        self.assertEqual(pretty_price("1.1"), "£1.10")
        self.assertEqual(pretty_price("1.11"), "£1.11")
        self.assertEqual(pretty_price("1.111"), "£1.11")
        self.assertEqual(pretty_price("123456789"), "£123,456,789")
        self.assertEqual(pretty_price("123456789.01"), "£123,456,789.01")

    def test_happy_not_simplified(self):
        self.assertEqual(pretty_price("0", False), "Free")
        self.assertEqual(pretty_price("0.1"), "£0.10")
        self.assertEqual(pretty_price("0.10"), "£0.10")
        self.assertEqual(pretty_price("0.101"), "£0.10")
        self.assertEqual(pretty_price("0.001"), "Free")
        self.assertEqual(pretty_price("0.009"), "£0.01")
        self.assertEqual(pretty_price("5", False), "£5.00")
        self.assertEqual(pretty_price("5.0", False), "£5.00")
        self.assertEqual(pretty_price("5.00", False), "£5.00")


class TestPrettyPriceRange(unittest.TestCase):
    def test_happy(self):
        self.assertEqual(pretty_price_range(0, 0), "Free")
        self.assertEqual(pretty_price_range("0", "0"), "Free")
        self.assertEqual(pretty_price_range(5, 5), "£5")
        self.assertEqual(pretty_price_range("5", 5), "£5")
        self.assertEqual(pretty_price_range(5, "5"), "£5")
        self.assertEqual(pretty_price_range("5", "5"), "£5")
        self.assertEqual(pretty_price_range(5.1, 5.1), "£5.10")
        self.assertEqual(pretty_price_range(0, 5), "Free to £5")
        self.assertEqual(pretty_price_range(None, 5), "Free to £5")
        self.assertEqual(pretty_price_range(5, 0), "From £5")
        self.assertEqual(pretty_price_range(5, 10), "£5 to £10")
        self.assertEqual(pretty_price_range(10, 5), "£5 to £10")
        with self.assertRaises(ValueError):
            pretty_price_range(5, "a")
        with self.assertRaises(TypeError):
            pretty_price_range(5, [])
        with self.assertRaises(TypeError):
            pretty_price_range(5, True)



def strtobool(value: str) -> bool:
    """
    Convert a string representation of truth to true (1) or false (0).

    Raises TypeError if 'val' is not a string.
    True values are 'y', 'yes', 't', 'true', 'on', and '1'.
    False values are 'n', 'no', 'f', 'false', 'off', and '0'.
    Raises ValueError if 'val' is anything else.
    """
    try:
        value = value.lower()
    except AttributeError:
        raise TypeError("invalid truth value %r" % (value,))

    if value in ("y", "yes", "t", "true", "on", "1"):
        return True
    if value in ("n", "no", "f", "false", "off", "0"):
        return False
    raise ValueError("invalid truth value %r" % (value,))



import re
import unicodedata


def slugify(value: str, allow_unicode=False) -> str:
    """
    Convert a string to a URL-friendly slug.

    Derived from Django
    https://github.com/django/django/blob/stable/6.0.x/django/utils/text.py#L469-L486

    Copyright (c) Django Software Foundation and individual contributors.
    All rights reserved.

    Redistribution and use in source and binary forms, with or without modification,
    are permitted provided that the following conditions are met:

        1. Redistributions of source code must retain the above copyright notice,
        this list of conditions and the following disclaimer.

        2. Redistributions in binary form must reproduce the above copyright
        notice, this list of conditions and the following disclaimer in the
        documentation and/or other materials provided with the distribution.

        3. Neither the name of Django nor the names of its contributors may be used
        to endorse or promote products derived from this software without
        specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
    ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
    WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
    ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
    (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
    ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
    SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
    """

    if not isinstance(value, str):
        raise TypeError("value must be a string")

    if not value:
        return value

    if allow_unicode:
        value = unicodedata.normalize("NFKC", value)
    else:
        value = (
            unicodedata.normalize("NFKD", value)
            .encode("ascii", "ignore")
            .decode("ascii")
        )
    value = re.sub(r"[^\w\s-]", "", value.lower())
    return re.sub(r"[-\s]+", "-", value).strip("-_")


def unslugify(slug: str, capitalize_first: bool = True) -> str:
    """
    Convert a slug back to a human-readable string.
    """

    if not isinstance(slug, str):
        raise TypeError("slug must be a string")

    if not slug:
        return slug

    string = slug.split("-")
    if capitalize_first:
        string[0] = string[0].capitalize()
    return " ".join(string)



import datetime
import unittest

from tna_utilities.datetime import (
    get_date_from_string,
    group_by_year_and_month,
    is_today_in_date_range,
    is_today_or_future,
    pretty_date,
    pretty_date_range,
    pretty_datetime,
    pretty_datetime_range,
    rfc_822_date_format,
    seconds_to_duration,
    seconds_to_iso_8601_duration,
)


class TestGetDateFromString(unittest.TestCase):
    def test_happy_dd_mm_yyyy(self):
        self.assertEqual(
            get_date_from_string("2006-05-04"),
            datetime.datetime(2006, 5, 4, 0, 0, 0),
        )

    def test_happy_mm_yyyy(self):
        self.assertEqual(
            get_date_from_string("2006-05"), datetime.datetime(2006, 5, 1, 0, 0, 0)
        )

    def test_happy_yyyy(self):
        self.assertEqual(
            get_date_from_string("2006"), datetime.datetime(2006, 1, 1, 0, 0, 0)
        )

    def test_happy_iso_8601(self):
        self.assertEqual(
            get_date_from_string("2006-05-04T01:02:03"),
            datetime.datetime(2006, 5, 4, 1, 2, 3),
        )

    def test_happy_iso_8601_microseconds(self):
        self.assertEqual(
            get_date_from_string("2006-05-04T01:02:03.999"),
            datetime.datetime(2006, 5, 4, 1, 2, 3, 999000),
        )

    def test_happy_iso_8601_timezone(self):
        self.assertEqual(
            get_date_from_string("2006-05-04T01:02:03+01:00"),
            datetime.datetime(
                2006,
                5,
                4,
                1,
                2,
                3,
                tzinfo=datetime.timezone(datetime.timedelta(hours=1)),
            ),
        )

    def test_happy_iso_8601_zulu(self):
        self.assertEqual(
            get_date_from_string("2006-05-04T01:02:03Z"),
            datetime.datetime(2006, 5, 4, 1, 2, 3, tzinfo=datetime.timezone.utc),
        )

    def test_happy_iso_8601_plain(self):
        self.assertEqual(
            get_date_from_string("1000"), datetime.datetime(1000, 1, 1, 0, 0, 0)
        )

    def test_unhappy_invalid_day(self):
        with self.assertRaises(ValueError):
            get_date_from_string("2006-12-32")

    def test_unhappy_invalid_month(self):
        with self.assertRaises(ValueError):
            get_date_from_string("2006-13")

    def test_unhappy_invalid_years(self):
        with self.assertRaises(ValueError):
            get_date_from_string("999")
        with self.assertRaises(ValueError):
            get_date_from_string("06")
        with self.assertRaises(ValueError):
            get_date_from_string("00")
        with self.assertRaises(ValueError):
            get_date_from_string("99")
        with self.assertRaises(ValueError):
            get_date_from_string("9")
        with self.assertRaises(ValueError):
            get_date_from_string("abc")

    def test_unhappy_blank_values(self):
        with self.assertRaises(ValueError):
            get_date_from_string("")
        with self.assertRaises(ValueError):
            get_date_from_string(None)
        with self.assertRaises(ValueError):
            get_date_from_string(False)


class TestPrettyDate(unittest.TestCase):
    def test_happy_string(self):
        self.assertEqual(pretty_date("2000-01-01T12:00:00Z"), "1 January 2000")
        self.assertEqual(pretty_date("2000-01-01"), "1 January 2000")
        self.assertEqual(pretty_date("2000-12-31"), "31 December 2000")
        self.assertEqual(pretty_date("2000-01"), "January 2000")
        self.assertEqual(pretty_date("2000"), "2000")

    def test_happy_string_show_day(self):
        self.assertEqual(
            pretty_date("2000-01-01T12:00:00Z", show_day=True),
            "Saturday 1 January 2000",
        )
        self.assertEqual(
            pretty_date("2000-01-01", show_day=True), "Saturday 1 January 2000"
        )
        self.assertEqual(
            pretty_date("2000-12-31", show_day=True), "Sunday 31 December 2000"
        )
        self.assertEqual(pretty_date("2000-01", show_day=True), "January 2000")
        self.assertEqual(pretty_date("2000", show_day=True), "2000")
        self.assertEqual(
            pretty_date("2000-01-01T12:30:00Z", show_day=True),
            "Saturday 1 January 2000",
        )

    def test_happy_date(self):
        date = datetime.date(2000, 1, 1)
        self.assertEqual(pretty_date(date), "1 January 2000")
        self.assertEqual(pretty_date(date, show_day=True), "Saturday 1 January 2000")

    def test_happy_datetime(self):
        date = datetime.datetime(2000, 1, 1, 12, 30, 0)
        self.assertEqual(pretty_date(date), "1 January 2000")
        self.assertEqual(pretty_date(date, show_day=True), "Saturday 1 January 2000")

    def test_unhappy_none(self):
        with self.assertRaises(ValueError):
            pretty_date(None)


class TestPrettyDatetime(unittest.TestCase):
    def test_happy_string(self):
        self.assertEqual(
            pretty_datetime("2000-01-01T12:00:00Z"), "1 January 2000, 12:00"
        )
        self.assertEqual(pretty_datetime("2000-01-01"), "1 January 2000, 00:00")
        self.assertEqual(pretty_datetime("2000-01"), "1 January 2000, 00:00")
        self.assertEqual(pretty_datetime("2000"), "1 January 2000, 00:00")

    def test_happy_string_show_day(self):
        self.assertEqual(
            pretty_datetime("2000-01-01T12:00:00Z", show_day=True),
            "Saturday 1 January 2000, 12:00",
        )
        self.assertEqual(
            pretty_datetime("2000-01-01", show_day=True),
            "Saturday 1 January 2000, 00:00",
        )
        self.assertEqual(
            pretty_datetime("2000-01", show_day=True), "Saturday 1 January 2000, 00:00"
        )
        self.assertEqual(
            pretty_datetime("2000", show_day=True), "Saturday 1 January 2000, 00:00"
        )

    def test_happy_datetime(self):
        date = datetime.datetime(2000, 1, 1, 12, 30, 0)
        self.assertEqual(pretty_datetime(date), "1 January 2000, 12:30")
        self.assertEqual(
            pretty_datetime(date, show_day=True), "Saturday 1 January 2000, 12:30"
        )

    def test_unhappy_date(self):
        date = datetime.date(2000, 1, 1)
        with self.assertRaises(TypeError):
            pretty_datetime(date)

    def test_unhappy_none(self):
        with self.assertRaises(ValueError):
            pretty_datetime(None)


class TestPrettyDateRange(unittest.TestCase):
    def test_happy_string(self):
        start_date = "2000-01-01"
        self.assertEqual(pretty_date_range(start_date, "2000-01-01"), "1 January 2000")
        self.assertEqual(
            pretty_date_range(start_date, "2000-01-02"), "1 to 2 January 2000"
        )
        self.assertEqual(
            pretty_date_range(start_date, "2000-01-31"), "1 to 31 January 2000"
        )
        self.assertEqual(
            pretty_date_range(start_date, "2000-02-01"),
            "1 January to 1 February 2000",
        )
        self.assertEqual(pretty_date_range(start_date, "2000-12-31"), "2000")
        self.assertEqual(
            pretty_date_range(start_date, "2001-01-01"),
            "1 January 2000 to 1 January 2001",
        )
        self.assertEqual(pretty_date_range(start_date, "2001-12-31"), "2000 to 2001")
        self.assertEqual(
            pretty_date_range(None, "2001-12-31"), "Now to 31 December 2001"
        )
        self.assertEqual(pretty_date_range(start_date, None), "From 1 January 2000")

    def test_happy_date(self):
        start_date = datetime.date(2000, 1, 1)
        self.assertEqual(
            pretty_date_range(start_date, datetime.date(2000, 1, 1)),
            "1 January 2000",
        )
        self.assertEqual(
            pretty_date_range(start_date, datetime.date(2000, 1, 2)),
            "1 to 2 January 2000",
        )
        self.assertEqual(
            pretty_date_range(start_date, datetime.date(2000, 1, 31)),
            "1 to 31 January 2000",
        )
        self.assertEqual(
            pretty_date_range(start_date, datetime.date(2000, 2, 1)),
            "1 January to 1 February 2000",
        )
        self.assertEqual(
            pretty_date_range(start_date, datetime.date(2000, 12, 31)), "2000"
        )
        self.assertEqual(
            pretty_date_range(start_date, datetime.date(2001, 1, 1)),
            "1 January 2000 to 1 January 2001",
        )
        self.assertEqual(
            pretty_date_range(start_date, datetime.date(2001, 12, 31)), "2000 to 2001"
        )
        self.assertEqual(
            pretty_date_range(None, datetime.date(2001, 12, 31)),
            "Now to 31 December 2001",
        )
        self.assertEqual(pretty_date_range(start_date, None), "From 1 January 2000")
        self.assertEqual(
            pretty_date_range(None, datetime.date(2001, 12, 31), lowercase_first=True),
            "now to 31 December 2001",
        )
        self.assertEqual(
            pretty_date_range(start_date, None, lowercase_first=True),
            "from 1 January 2000",
        )

    def test_happy_datetime(self):
        start_date = datetime.datetime(2000, 1, 1)
        self.assertEqual(
            pretty_date_range(start_date, datetime.datetime(2000, 1, 1, 14, 45, 0)),
            "1 January 2000",
        )
        self.assertEqual(
            pretty_date_range(start_date, datetime.datetime(2000, 1, 2, 14, 45, 0)),
            "1 to 2 January 2000",
        )
        self.assertEqual(
            pretty_date_range(start_date, datetime.datetime(2000, 1, 31, 14, 45, 0)),
            "1 to 31 January 2000",
        )
        self.assertEqual(
            pretty_date_range(start_date, datetime.datetime(2000, 2, 1, 14, 45, 0)),
            "1 January to 1 February 2000",
        )
        self.assertEqual(
            pretty_date_range(start_date, datetime.datetime(2000, 12, 31, 14, 45, 0)),
            "2000",
        )
        self.assertEqual(
            pretty_date_range(start_date, datetime.datetime(2001, 1, 1, 14, 45, 0)),
            "1 January 2000 to 1 January 2001",
        )
        self.assertEqual(
            pretty_date_range(start_date, datetime.datetime(2001, 12, 31, 14, 45, 0)),
            "2000 to 2001",
        )
        self.assertEqual(
            pretty_date_range(None, datetime.datetime(2001, 12, 31, 14, 45, 0)),
            "Now to 31 December 2001",
        )
        self.assertEqual(pretty_date_range(start_date, None), "From 1 January 2000")
        self.assertEqual(
            pretty_date_range(
                None, datetime.datetime(2001, 12, 31, 14, 45, 0), lowercase_first=True
            ),
            "now to 31 December 2001",
        )
        self.assertEqual(
            pretty_date_range(start_date, None, lowercase_first=True),
            "from 1 January 2000",
        )

    def test_happy_pretty_date_range_no_days(self):
        start_date = datetime.date(2000, 1, 1)
        self.assertEqual(
            pretty_date_range(start_date, datetime.date(2000, 1, 1), omit_days=True),
            "January 2000",
        )
        self.assertEqual(
            pretty_date_range(start_date, datetime.date(2000, 1, 2), omit_days=True),
            "January 2000",
        )
        self.assertEqual(
            pretty_date_range(start_date, datetime.date(2000, 1, 31), omit_days=True),
            "January 2000",
        )
        self.assertEqual(
            pretty_date_range(start_date, datetime.date(2000, 2, 1), omit_days=True),
            "January to February 2000",
        )
        self.assertEqual(
            pretty_date_range(start_date, datetime.date(2000, 12, 31), omit_days=True),
            "2000",
        )
        self.assertEqual(
            pretty_date_range(start_date, datetime.date(2001, 1, 1), omit_days=True),
            "January 2000 to January 2001",
        )
        self.assertEqual(
            pretty_date_range(start_date, datetime.date(2001, 12, 31), omit_days=True),
            "2000 to 2001",
        )
        self.assertEqual(
            pretty_date_range(start_date, None, omit_days=True),
            "From January 2000",
        )
        self.assertEqual(
            pretty_date_range(None, datetime.date(2001, 12, 31), omit_days=True),
            "Now to December 2001",
        )
        self.assertEqual(
            pretty_date_range(start_date, None, omit_days=True, lowercase_first=True),
            "from January 2000",
        )
        self.assertEqual(
            pretty_date_range(
                None, datetime.date(2001, 12, 31), omit_days=True, lowercase_first=True
            ),
            "now to December 2001",
        )

    def test_unhappy_order(self):
        with self.assertRaises(ValueError):
            pretty_date_range(
                datetime.datetime(2001, 1, 1), datetime.datetime(2000, 1, 1)
            )

    def test_unhappy_none(self):
        with self.assertRaises(ValueError):
            pretty_date_range(None, None)


class TestPrettyDatetimeRange(unittest.TestCase):
    def test_happy_string(self):
        start_date = "2000-01-01T12:30:00Z"
        self.assertEqual(
            pretty_datetime_range(start_date, "2000-01-01T12:30:00Z"),
            "1 January 2000, 12:30",
        )
        self.assertEqual(
            pretty_datetime_range(start_date, "2000-01-01T12:31:00Z"),
            "1 January 2000, 12:30 to 12:31",
        )
        self.assertEqual(
            pretty_datetime_range(start_date, "2000-01-01T23:59:59Z"),
            "1 January 2000, 12:30 to 23:59",
        )
        self.assertEqual(
            pretty_datetime_range(start_date, "2000-01-02T00:00:00Z"),
            "1 January 2000, 12:30 to 2 January 2000, 00:00",
        )
        self.assertEqual(
            pretty_datetime_range(start_date, "2000-01-02T14:45:00Z"),
            "1 January 2000, 12:30 to 2 January 2000, 14:45",
        )
        self.assertEqual(
            pretty_datetime_range(start_date, "2000-01-31T14:45:00Z"),
            "1 January 2000, 12:30 to 31 January 2000, 14:45",
        )
        self.assertEqual(
            pretty_datetime_range(start_date, "2000-02-01T14:45:00Z"),
            "1 January 2000, 12:30 to 1 February 2000, 14:45",
        )
        self.assertEqual(
            pretty_datetime_range(start_date, "2000-12-31T14:45:00Z"),
            "1 January 2000, 12:30 to 31 December 2000, 14:45",
        )
        self.assertEqual(
            pretty_datetime_range(start_date, "2001-01-01T14:45:00Z"),
            "1 January 2000, 12:30 to 1 January 2001, 14:45",
        )
        self.assertEqual(
            pretty_datetime_range(start_date, "2001-12-31T14:45:00Z"),
            "1 January 2000, 12:30 to 31 December 2001, 14:45",
        )
        self.assertEqual(
            pretty_datetime_range(start_date, None),
            "From 1 January 2000, 12:30",
        )
        self.assertEqual(
            pretty_datetime_range(None, "2001-12-31T14:45:00Z"),
            "Now to 31 December 2001, 14:45",
        )
        self.assertEqual(
            pretty_datetime_range(start_date, None, lowercase_first=True),
            "from 1 January 2000, 12:30",
        )
        self.assertEqual(
            pretty_datetime_range(None, "2001-12-31T14:45:00Z", lowercase_first=True),
            "now to 31 December 2001, 14:45",
        )
        self.assertEqual(
            pretty_datetime_range(
                start_date, "2000-01-01T12:30:00Z", hide_date_if_single_day=True
            ),
            "12:30",
        )
        self.assertEqual(
            pretty_datetime_range(
                start_date, "2000-01-01T12:45:00Z", hide_date_if_single_day=True
            ),
            "12:30 to 12:45",
        )
        self.assertEqual(
            pretty_datetime_range(
                start_date, "2000-01-02T00:00:00Z", hide_date_if_single_day=True
            ),
            "1 January 2000, 12:30 to 2 January 2000, 00:00",
        )

    def test_happy_datetime(self):
        start_date = datetime.datetime(2000, 1, 1, 12, 30, 0)
        self.assertEqual(
            pretty_datetime_range(start_date, datetime.datetime(2000, 1, 1, 12, 30, 0)),
            "1 January 2000, 12:30",
        )
        self.assertEqual(
            pretty_datetime_range(start_date, datetime.datetime(2000, 1, 1, 12, 31, 0)),
            "1 January 2000, 12:30 to 12:31",
        )
        self.assertEqual(
            pretty_datetime_range(
                start_date, datetime.datetime(2000, 1, 1, 23, 59, 59)
            ),
            "1 January 2000, 12:30 to 23:59",
        )
        self.assertEqual(
            pretty_datetime_range(start_date, datetime.datetime(2000, 1, 2, 0, 0, 0)),
            "1 January 2000, 12:30 to 2 January 2000, 00:00",
        )
        self.assertEqual(
            pretty_datetime_range(start_date, datetime.datetime(2000, 1, 2, 14, 45, 0)),
            "1 January 2000, 12:30 to 2 January 2000, 14:45",
        )
        self.assertEqual(
            pretty_datetime_range(
                start_date, datetime.datetime(2000, 1, 31, 14, 45, 0)
            ),
            "1 January 2000, 12:30 to 31 January 2000, 14:45",
        )
        self.assertEqual(
            pretty_datetime_range(start_date, datetime.datetime(2000, 2, 1, 14, 45, 0)),
            "1 January 2000, 12:30 to 1 February 2000, 14:45",
        )
        self.assertEqual(
            pretty_datetime_range(
                start_date, datetime.datetime(2000, 12, 31, 14, 45, 0)
            ),
            "1 January 2000, 12:30 to 31 December 2000, 14:45",
        )
        self.assertEqual(
            pretty_datetime_range(start_date, datetime.datetime(2001, 1, 1, 14, 45, 0)),
            "1 January 2000, 12:30 to 1 January 2001, 14:45",
        )
        self.assertEqual(
            pretty_datetime_range(
                start_date, datetime.datetime(2001, 12, 31, 14, 45, 0)
            ),
            "1 January 2000, 12:30 to 31 December 2001, 14:45",
        )
        self.assertEqual(
            pretty_datetime_range(start_date, None),
            "From 1 January 2000, 12:30",
        )
        self.assertEqual(
            pretty_datetime_range(None, datetime.datetime(2001, 12, 31, 14, 45, 0)),
            "Now to 31 December 2001, 14:45",
        )
        self.assertEqual(
            pretty_datetime_range(start_date, None, lowercase_first=True),
            "from 1 January 2000, 12:30",
        )
        self.assertEqual(
            pretty_datetime_range(
                None, datetime.datetime(2001, 12, 31, 14, 45, 0), lowercase_first=True
            ),
            "now to 31 December 2001, 14:45",
        )
        self.assertEqual(
            pretty_datetime_range(
                start_date,
                datetime.datetime(2000, 1, 1, 12, 30, 0),
                hide_date_if_single_day=True,
            ),
            "12:30",
        )
        self.assertEqual(
            pretty_datetime_range(
                start_date,
                datetime.datetime(2000, 1, 1, 12, 45, 0),
                hide_date_if_single_day=True,
            ),
            "12:30 to 12:45",
        )
        self.assertEqual(
            pretty_datetime_range(
                start_date,
                datetime.datetime(2000, 1, 2, 0, 0, 0),
                hide_date_if_single_day=True,
            ),
            "1 January 2000, 12:30 to 2 January 2000, 00:00",
        )

    def test_unhappy_date(self):
        with self.assertRaises(TypeError):
            pretty_datetime_range(datetime.date(2000, 1, 1), None)
        with self.assertRaises(TypeError):
            pretty_datetime_range(None, datetime.date(2000, 1, 1))

    def test_unhappy_order(self):
        with self.assertRaises(ValueError):
            pretty_datetime_range(
                datetime.datetime(2001, 1, 1), datetime.datetime(2000, 1, 1)
            )

    def test_unhappy_none(self):
        with self.assertRaises(ValueError):
            pretty_datetime_range(None, None)


class TestIsTodayOrFuture(unittest.TestCase):
    def test_happy(self):
        self.assertTrue(is_today_or_future(datetime.date(2999, 1, 1)))
        self.assertFalse(is_today_or_future(datetime.date(2000, 1, 1)))
        today = datetime.datetime.now()
        self.assertTrue(is_today_or_future(today))
        tomorrow = today + datetime.timedelta(days=1)
        self.assertTrue(is_today_or_future(tomorrow))
        yesterday = today + datetime.timedelta(days=-1)
        self.assertFalse(is_today_or_future(yesterday))

    def test_unhappy(self):
        with self.assertRaises(ValueError):
            is_today_or_future(None)


class TestIsTodayInDateRange(unittest.TestCase):
    def test_happy_date(self):
        self.assertFalse(
            is_today_in_date_range(datetime.date(2000, 1, 1), datetime.date(2001, 1, 1))
        )
        self.assertTrue(
            is_today_in_date_range(datetime.date(2000, 1, 1), datetime.date(2999, 1, 1))
        )
        self.assertFalse(
            is_today_in_date_range(datetime.date(2998, 1, 1), datetime.date(2999, 1, 1))
        )

    def test_unhappy_string(self):
        with self.assertRaises(ValueError):
            is_today_in_date_range(None, "2023-10-31")
        with self.assertRaises(ValueError):
            is_today_in_date_range("2023-10-01", None)
        with self.assertRaises(ValueError):
            is_today_in_date_range(None, None)

    def test_unhappy_invalid(self):
        with self.assertRaises(ValueError):
            is_today_in_date_range(None, "foo")
        with self.assertRaises(ValueError):
            is_today_in_date_range("bar", None)

    def test_unhappy_date(self):
        with self.assertRaises(ValueError):
            is_today_in_date_range(None, datetime.date(2023, 10, 31))
        with self.assertRaises(ValueError):
            is_today_in_date_range(datetime.date(2023, 10, 31), None)
        with self.assertRaises(ValueError):
            is_today_in_date_range(None, None)


class TestGroupByYearAndMonth(unittest.TestCase):
    def test_happy_strings(self):
        self.maxDiff = None
        input_data = [
            {"id": 1, "date": "2022-05-15"},
            {"id": 2, "date": "2022-05-20"},
            {"id": 3, "date": "2022-06-10"},
            {"id": 4, "date": "2021-12-25"},
            {"id": 5, "date": "2021-11-11"},
            {"id": 6, "date": "2022-06-15"},
            {"id": 7, "date": "foobar"},
        ]
        result = group_by_year_and_month({"items": input_data}, "date")
        expected = [
            {
                "heading": "2021",
                "index": 2021,
                "items": [
                    {
                        "heading": "November",
                        "index": 11,
                        "items": [{"id": 5, "date": "2021-11-11"}],
                    },
                    {
                        "heading": "December",
                        "index": 12,
                        "items": [{"id": 4, "date": "2021-12-25"}],
                    },
                ],
            },
            {
                "heading": "2022",
                "index": 2022,
                "items": [
                    {
                        "heading": "May",
                        "index": 5,
                        "items": [
                            {"id": 1, "date": "2022-05-15"},
                            {"id": 2, "date": "2022-05-20"},
                        ],
                    },
                    {
                        "heading": "June",
                        "index": 6,
                        "items": [
                            {"id": 3, "date": "2022-06-10"},
                            {"id": 6, "date": "2022-06-15"},
                        ],
                    },
                ],
            },
        ]
        self.assertEqual(result, expected)

    def test_happy_datetime(self):
        input_data = [
            {"id": 1, "date": datetime.date(2022, 5, 15)},
            {"id": 2, "date": datetime.date(2022, 5, 20)},
            {"id": 3, "date": datetime.date(2022, 6, 10)},
            {"id": 4, "date": datetime.date(2021, 12, 25)},
            {"id": 5, "date": datetime.date(2021, 11, 11)},
            {"id": 6, "date": datetime.date(2022, 6, 15)},
            {"id": 7, "date": None},
        ]
        result = group_by_year_and_month({"items": input_data}, "date")
        expected = [
            {
                "heading": "2021",
                "index": 2021,
                "items": [
                    {
                        "heading": "November",
                        "index": 11,
                        "items": [{"id": 5, "date": datetime.date(2021, 11, 11)}],
                    },
                    {
                        "heading": "December",
                        "index": 12,
                        "items": [{"id": 4, "date": datetime.date(2021, 12, 25)}],
                    },
                ],
            },
            {
                "heading": "2022",
                "index": 2022,
                "items": [
                    {
                        "heading": "May",
                        "index": 5,
                        "items": [
                            {"id": 1, "date": datetime.date(2022, 5, 15)},
                            {"id": 2, "date": datetime.date(2022, 5, 20)},
                        ],
                    },
                    {
                        "heading": "June",
                        "index": 6,
                        "items": [
                            {"id": 3, "date": datetime.date(2022, 6, 10)},
                            {"id": 6, "date": datetime.date(2022, 6, 15)},
                        ],
                    },
                ],
            },
        ]
        self.assertEqual(result, expected)

    def test_happy_strings_reverse(self):
        self.maxDiff = None
        input_data = [
            {"id": 1, "date": "2022-05-15"},
            {"id": 2, "date": "2022-05-20"},
            {"id": 3, "date": "2022-06-10"},
            {"id": 4, "date": "2021-12-25"},
            {"id": 5, "date": "2021-11-11"},
            {"id": 6, "date": "2022-06-15"},
            {"id": 7, "date": "foobar"},
        ]
        result = group_by_year_and_month({"items": input_data}, "date", reverse=True)
        expected = [
            {
                "heading": "2022",
                "index": 2022,
                "items": [
                    {
                        "heading": "June",
                        "index": 6,
                        "items": [
                            {"id": 6, "date": "2022-06-15"},
                            {"id": 3, "date": "2022-06-10"},
                        ],
                    },
                    {
                        "heading": "May",
                        "index": 5,
                        "items": [
                            {"id": 2, "date": "2022-05-20"},
                            {"id": 1, "date": "2022-05-15"},
                        ],
                    },
                ],
            },
            {
                "heading": "2021",
                "index": 2021,
                "items": [
                    {
                        "heading": "December",
                        "index": 12,
                        "items": [{"id": 4, "date": "2021-12-25"}],
                    },
                    {
                        "heading": "November",
                        "index": 11,
                        "items": [{"id": 5, "date": "2021-11-11"}],
                    },
                ],
            },
        ]
        self.assertEqual(result, expected)

    def test_happy_datetime_reverse(self):
        input_data = [
            {"id": 1, "date": datetime.date(2022, 5, 15)},
            {"id": 2, "date": datetime.date(2022, 5, 20)},
            {"id": 3, "date": datetime.date(2022, 6, 10)},
            {"id": 4, "date": datetime.date(2021, 12, 25)},
            {"id": 5, "date": datetime.date(2021, 11, 11)},
            {"id": 6, "date": datetime.date(2022, 6, 15)},
            {"id": 7, "date": None},
        ]
        result = group_by_year_and_month({"items": input_data}, "date", reverse=True)
        expected = [
            {
                "heading": "2022",
                "index": 2022,
                "items": [
                    {
                        "heading": "June",
                        "index": 6,
                        "items": [
                            {"id": 6, "date": datetime.date(2022, 6, 15)},
                            {"id": 3, "date": datetime.date(2022, 6, 10)},
                        ],
                    },
                    {
                        "heading": "May",
                        "index": 5,
                        "items": [
                            {"id": 2, "date": datetime.date(2022, 5, 20)},
                            {"id": 1, "date": datetime.date(2022, 5, 15)},
                        ],
                    },
                ],
            },
            {
                "heading": "2021",
                "index": 2021,
                "items": [
                    {
                        "heading": "December",
                        "index": 12,
                        "items": [{"id": 4, "date": datetime.date(2021, 12, 25)}],
                    },
                    {
                        "heading": "November",
                        "index": 11,
                        "items": [{"id": 5, "date": datetime.date(2021, 11, 11)}],
                    },
                ],
            },
        ]
        self.assertEqual(result, expected)


class TestSecondsToIso8601Duration(unittest.TestCase):
    def test_happy(self):
        self.assertEqual(seconds_to_iso_8601_duration(0), "PT0S")
        self.assertEqual(seconds_to_iso_8601_duration(1), "PT1S")
        self.assertEqual(seconds_to_iso_8601_duration(59), "PT59S")
        self.assertEqual(seconds_to_iso_8601_duration(60), "PT1M0S")
        self.assertEqual(seconds_to_iso_8601_duration(61), "PT1M1S")
        self.assertEqual(seconds_to_iso_8601_duration(1337), "PT22M17S")
        self.assertEqual(seconds_to_iso_8601_duration(3599), "PT59M59S")
        self.assertEqual(seconds_to_iso_8601_duration(3600), "PT1H0M0S")
        self.assertEqual(seconds_to_iso_8601_duration(3601), "PT1H0M1S")

    def test_unhappy_negative(self):
        with self.assertRaises(ValueError):
            seconds_to_iso_8601_duration(-1)


class TestSecondsToTime(unittest.TestCase):
    def test_happy(self):
        self.assertEqual(seconds_to_duration(0), "00h 00m 00s")
        self.assertEqual(seconds_to_duration(1), "00h 00m 01s")
        self.assertEqual(seconds_to_duration(59), "00h 00m 59s")
        self.assertEqual(seconds_to_duration(60), "00h 01m 00s")
        self.assertEqual(seconds_to_duration(61), "00h 01m 01s")
        self.assertEqual(seconds_to_duration(1337), "00h 22m 17s")
        self.assertEqual(seconds_to_duration(3599), "00h 59m 59s")
        self.assertEqual(seconds_to_duration(3600), "01h 00m 00s")
        self.assertEqual(seconds_to_duration(3601), "01h 00m 01s")

    def test_happy_simplified(self):
        self.assertEqual(seconds_to_duration(0, simplify=True), "00s")
        self.assertEqual(seconds_to_duration(1, simplify=True), "01s")
        self.assertEqual(seconds_to_duration(59, simplify=True), "59s")
        self.assertEqual(seconds_to_duration(60, simplify=True), "01m 00s")
        self.assertEqual(seconds_to_duration(61, simplify=True), "01m 01s")
        self.assertEqual(seconds_to_duration(1337, simplify=True), "22m 17s")
        self.assertEqual(seconds_to_duration(3599, simplify=True), "59m 59s")
        self.assertEqual(seconds_to_duration(3600, simplify=True), "01h 00m 00s")
        self.assertEqual(seconds_to_duration(3601, simplify=True), "01h 00m 01s")

    def test_unhappy_negative(self):
        with self.assertRaises(ValueError):
            seconds_to_duration(-1)


class TestRfc822DateFormat(unittest.TestCase):
    def test_happy(self):
        self.assertEqual(
            rfc_822_date_format(datetime.datetime(2000, 1, 1, 12, 30, 45)),
            "Sat, 1 Jan 2000 12:30:45 GMT",
        )



from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="tna-utilities",
    version="1.1.0",
    author="Andrew Hosgood",
    author_email="andrew.hosgood@nationalarchives.gov.uk",
    description="A library of common National Archives Python functions",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nationalarchives/python-utilities",
    project_urls={
        "Bug Tracker": "https://github.com/nationalarchives/python-utilities/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
    ],
    packages=find_packages(include=["tna_utilities"], exclude=["test"]),
    python_requires=">=3.10",
)
